from axiom import attributes, item
from axiom.errors import ItemNotFound
from twisted.python.reflect import namedAny


class _PersistedFactory(item.Item):
    identifier = attributes.bytes(indexed=True)
    name = attributes.bytes()

    def dereference(self):
        return namedAny(self.name)



class FactoryDict(object):
    """A collection of multiplexable factories.

    This object is read-only; to persist factories, use
    ``addToStore``.

    """
    def __init__(self, store):
        self.store = store


    def __getitem__(self, key):
        PF = _PersistedFactory
        try:
            persistedFactory = self.store.findUnique(PF, PF.identifier == key)
        except ItemNotFound:
            raise KeyError(key)

        return persistedFactory.dereference()



def addToStore(store, identifier, name):
    """Adds a persisted factory with given identifier and object name to
    the given store.

    This is intended to have the identifier and name partially
    applied, so that a particular module with an exercise in it can
    just have an ``addToStore`` function that remembers it in the
    store.

    The only difference between this and ``_PeristedFactory`` itself is
    that you have to pass a store to ``_PersistedFactory`` as the
    ``store`` keyword argument, whereas you can pass it to this
    function positionally.

    """
    return _PersistedFactory(store=store, identifier=identifier, name=name)



from axiom.store import Store
from functools import partial
from merlyn.multiplexing import addToStore, FactoryDict, _PersistedFactory
from twisted.trial.unittest import SynchronousTestCase


theFactory = object()
theNameOfTheFactory = "merlyn.test.test_multiplexing.theFactory"


class FactoryDictTests(SynchronousTestCase):
    """FactoryDicts are like dicts of names and identifiers except in an
    axiom store.

    """
    def setUp(self):
        self.store = Store()
        self.factoryDict = FactoryDict(self.store)
        _PersistedFactory(store=self.store,
                          identifier="identifier",
                          name=theNameOfTheFactory)


    def test_get(self):
        """FactoryDicts supply factories by identifiers.

        """
        self.assertIdentical(self.factoryDict["identifier"], theFactory)


    def test_getMissing(self):
        """FactoryDicts raise KeyError when there is no factory with the given
        identifier.

        """
        self.assertRaises(KeyError, lambda: self.factoryDict["BOGUS"])


    def test_addToStore(self):
        """The addToStore function adds a persisted factory to a store.

        """
        addThisToStore = partial(addToStore,
                                 identifier="otherIdentifier",
                                 name=theNameOfTheFactory)

        self.assertRaises(KeyError, lambda: self.factoryDict["otherIdentifier"])
        addThisToStore(self.store)
        self.assertIdentical(self.factoryDict["otherIdentifier"], theFactory)



from axiom.store import Store
from clarent.exercise import GetExercises, GetExerciseDetails
from clarent.exercise import UnknownExercise, NotifySolved
from merlyn.exercise import Exercise, Locator, solveAndNotify
from merlyn.auth import User
from twisted.trial.unittest import SynchronousTestCase
from txampext.respondertests import ResponderTestMixin



defaultExerciseKwargs = {
    "identifier": b"identifier",
    "title": u"\N{SNOWMAN}",
    "description": u"\N{CLOUD}"
}


def makeExercise(**kwargs):
    kw = defaultExerciseKwargs.copy()
    kw.update(kwargs)
    return Exercise(**kw)



class ExerciseTests(SynchronousTestCase):
    def test_wasSolvedBy(self):
        """The exercise knows when it was solved by a particular user.

        """
        store = Store()
        exercise = makeExercise(store=store)

        someUser = User(store=store, email="foo@example.com")
        self.assertFalse(exercise.wasSolvedBy(someUser))

        exercise.solvedBy(someUser)
        self.assertTrue(exercise.wasSolvedBy(someUser))

        someOtherUser = User(store=store, email="bar@example.com")
        self.assertFalse(exercise.wasSolvedBy(someOtherUser))



class SolveAndNotifyTests(SynchronousTestCase):
    def test_solveAndNotify(self):
        store = Store()
        user = User(store=store, email="test@example.com")
        proto = FakeProto(store, user)
        exercise = makeExercise(store=store)

        self.assertFalse(exercise.wasSolvedBy(user))
        solveAndNotify(proto, exercise)
        self.assertTrue(exercise.wasSolvedBy(user))

        self.assertIdentical(proto.command, NotifySolved)
        self.assertEqual(proto.kwargs, {
            b"identifier": b"identifier",
            b"title": u"\N{SNOWMAN}"
        })


class FakeProto(object):
    def __init__(self, store, user):
        self.store = store
        self.user = user

        self.command = self.kwargs = None


    def callRemote(self, command, **kwargs):
        self.command = command
        self.kwargs = kwargs



class _LocatorTests(object):
    def setUp(self):
        self.locator = Locator()
        self.locator.store = store = Store()
        self.locator.user = user = User(store=store, email=b"x@y.z")

        one = makeExercise(store=store, title=u"Exercise 1", identifier=b"1")
        one.solvedBy(user)

        makeExercise(store=store, title=u"Exercise 2", identifier=b"2")
        makeExercise(store=store, title=u"Exercise 3", identifier=b"3")



class GetExercisesTests(_LocatorTests, SynchronousTestCase):
    def test_getExercises(self):
        """A user can get some new exercises.

        """
        response = self.locator.getExercises(solved=False)
        exercises = list(response["exercises"])
        exercises.sort(key=lambda d: d[b"identifier"])
        self.assertEqual(exercises, [
            {b"title": u"Exercise 2", b"identifier": b"2"},
            {b"title": u"Exercise 3", b"identifier": b"3"}
        ])


    def test_getSolvedExercises(self):
        """A user can get previously solved exercises.

        """
        response = self.locator.getExercises(solved=True)
        exercises = list(response["exercises"])
        exercises.sort(key=lambda d: d["identifier"])
        self.assertEqual(exercises, [
            {b"title": u"Exercise 1", b"identifier": b"1"},
        ])



class GetExerciseDetailsTests(_LocatorTests, SynchronousTestCase):
    def test_getSolvedExerciseDetails(self):
        """A user can get details about a solved exercise.

        """
        details = self.locator.getExerciseDetails(identifier=b"1")
        self.assertEqual(details, {
            b"identifier": b"1",
            b"title": u"Exercise 1",
            b"description": u"\N{CLOUD}",
            b"solved": True
        })


    def test_getUnsolvedExerciseDetails(self):
        """A user can get details about an unsolved exercise.

        """
        details = self.locator.getExerciseDetails(identifier=b"2")
        self.assertEqual(details, {
            b"identifier": "2",
            b"title": u"Exercise 2",
            b"description": u"\N{CLOUD}",
            b"solved": False
        })


    def test_missingExercise(self):
        """When attempting to get details for an exercise that doesn't exist,
        an error is raised.

        """
        self.assertRaises(UnknownExercise,
                          self.locator.getExerciseDetails, identifier="BOGUS")



locator = Locator()



class GetExercisesResponderTests(ResponderTestMixin, SynchronousTestCase):
    command =  GetExercises
    locator = locator



class GetExerciseDetailsResponderTests(ResponderTestMixin, SynchronousTestCase):
    command =  GetExerciseDetails
    locator = locator



from axiom import store
from merlyn import auth, exercise, manhole
from twisted.application import service
from twisted import plugin
from twisted.internet import reactor
from twisted.internet.protocol import ServerFactory
from twisted.protocols.amp import AMP
from twisted.python import usage
from zope import interface


class AMP(AMP, exercise.Locator, auth.UserMixin):
    """
    The merlyn AMP protocol.
    """
    @property
    def store(self):
        return self.factory.store


    def connectionMade(self):
        """Keep a reference to the protocol on the factory.

        Unfortunately, we can't add the protocol by TLS certificate
        fingerprint, because the TLS handshake won't have completed
        yet, so ``self.transport.getPeerCertificate()`` is still
        ``None``.

        """
        self.factory.protocols.add(self)
        super(AMP, self).connectionMade()


    def connectionLost(self, reason):
        """Lose the reference to the protocol on the factory.

        """
        self.factory.protocols.remove(self)
        super(AMP, self).connectionLost(reason)



class AMPFactory(ServerFactory):
    protocol = AMP

    def __init__(self, store):
        self.store = store
        self.protocols = set()



class Options(usage.Options):
    """
    The options for starting a service.
    """
    optParameters = [
        ["store", "s", None, "Path to the store (mandatory)", store.Store]
    ]

    def postOptions(self):
        """Verify that a store was passed.

        """
        if self["store"] is None:
            raise usage.UsageError("Passing a root store is mandatory.")



class Service(service.Service):
    def __init__(self, store, reactor=reactor):
        self.store = store
        self.reactor = reactor


    def startService(self):
        factory = AMPFactory(store)
        ctxFactory = auth.ContextFactory(self.store)
        self.reactor.listenSSL(4430, factory, ctxFactory)

        manholeFactory = manhole.makeFactory(self.store, factory)
        self.reactor.listenTCP(8888, manholeFactory, interface="localhost")



@interface.implementer(plugin.IPlugin, service.IServiceMaker)
class ServiceMaker(object):
    """
    Makes a merlyn service from the command line.
    """
    tapname = "merlyn"
    description = "Exercise server"
    options = Options

    def makeService(self, options):
        return Service(options["store"])



"""Note: this is horribly insecure; only listen on localhost ever!

"""
from twisted.cred import checkers, portal
from twisted.conch import manhole, manhole_ssh

checker = checkers.InMemoryUsernamePasswordDatabaseDontUse(manhole="manhole")


def makeFactory(store, ampFactory):
    shellGlobals = {"store": store, "ampFactory": ampFactory}
    makeManhole = lambda _ign: manhole.ColoredManhole(shellGlobals)

    realm = manhole_ssh.TerminalRealm()
    realm.chainedProtocolFactory.protocolFactory = makeManhole

    p = portal.Portal(realm, [checker])
    return manhole_ssh.ConchFactory(p)



import sys
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
from setuptools.command import egg_info

packageName = "merlyn"

import re
versionLine = open("{0}/_version.py".format(packageName), "rt").read()
match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", versionLine, re.M)
versionString = match.group(1)

class Tox(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_suite = True

    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import tox
        sys.exit(tox.cmdline([]))

def _topLevel(name):
    return name.split('.', 1)[0]

def _hacked_write_toplevel_names(cmd, basename, filename):
    names = map(_topLevel, cmd.distribution.iter_distribution_names())
    pkgs = dict.fromkeys(set(names) - set(["twisted"]))
    cmd.write_file("top-level names", filename, '\n'.join(pkgs) + '\n')

egg_info.write_toplevel_names = _hacked_write_toplevel_names

setup(name=packageName,
      version=versionString,
      description='A server backend for interactive online exercises.',
      long_description=open("README.rst").read(),
      url='https://github.com/crypto101/' + packageName,

      author='Laurens Van Houtven',
      author_email='_@lvh.io',

      packages=find_packages() + ['twisted.plugins'],
      test_suite=packageName + ".test",
      cmdclass={'test': Tox},
      zip_safe=True,

      license='ISC',
      keywords="crypto twisted",
      classifiers=[
          "Development Status :: 3 - Alpha",
          "Framework :: Twisted",
          "Intended Audience :: Education",
          "License :: OSI Approved :: ISC License (ISCL)",
          "Programming Language :: Python :: 2 :: Only",
          "Programming Language :: Python :: 2.6",
          "Programming Language :: Python :: 2.7",
          "Topic :: Education",
          "Topic :: Games/Entertainment",
          "Topic :: Security :: Cryptography",
        ]
)

try:
    from twisted.plugin import IPlugin, getPlugins
except ImportError:
    pass
else:
    list(getPlugins(IPlugin))
